# -*- coding: utf-8 -*-
"""
Created on Mon Dec 11 14:03:23 2023

Example and test for using umbridge with bayesvalidrox.
Example model is the tsunami model by 
docker run -it -p 4242:4242 -v ~/tsunami_output:/output linusseelinger/model-exahype-tsunami

@author: rkohlhaas
"""

import joblib
import numpy as np 
import umbridge

import bayesvalidrox as bv


if __name__ == '__main__':
    
    # This model has 2 inputs and four outputs
    n_prior_sample = 1000
    priors = bv.Input()
    priors.add_marginals()
    priors.Marginals[0].name = 'x'
    priors.Marginals[0].input_data = np.random.uniform(50,150,n_prior_sample)  
    priors.add_marginals()
    priors.Marginals[1].name = 'y'
    priors.Marginals[1].input_data = np.random.uniform(50,150,n_prior_sample)
    
    # Define the model - level 0
    model0 = bv.PyLinkForwardModel()
    #model.link_type = 'Function'
    model0.link_type = 'umbridge'
    model0.py_file = 'tsunami_model'
    model0.name = 'tsunami_model'
    model0.Output.names = ['T1', 'T2', 'H1', 'H2']
    #model.observations = data_dict1_times
    
    # Define the model - level 1
    model1 = bv.PyLinkForwardModel()
    #model.link_type = 'Function'
    model1.link_type = 'umbridge'
    model1.py_file = 'tsunami_model1'
    model1.name = 'tsunami_model1'
    model1.Output.names = ['T1', 'T2', 'H1', 'H2']
    #model.observations = data_dict1_times
    
    # Create the surrogate
    surrogate_opts0 = bv.MetaModel(priors, model0)
    
    # Select your metamodel method
    surrogate_opts0.meta_model_type = 'aPCE'
    surrogate_opts0.pce_reg_method = 'FastARD'
    surrogate_opts0.pce_deg = np.arange(1, 5)
    surrogate_opts0.pce_q_norm = 0.4#1.0

    # Select your metamodel method         
    surrogate_opts0.add_ExpDesign()
    surrogate_opts0.ExpDesign.method = 'normal'
    surrogate_opts0.ExpDesign.n_init_samples = 50
    surrogate_opts0.ExpDesign.sampling_method = 'latin-hypercube'
    surrogate0 = surrogate_opts0.create_metamodel()
    print('Surrogate 0 completed')
    print('')
    
    # Create the surrogate
    surrogate_opts1 = bv.MetaModel(priors, model1)
    
    # Select your metamodel method
    surrogate_opts1.meta_model_type = 'aPCE'
    surrogate_opts1.pce_reg_method = 'FastARD'
    surrogate_opts1.pce_deg = np.arange(1, 5)
    surrogate_opts1.pce_q_norm = 0.4#1.0

    # Select your metamodel method         
    surrogate_opts1.add_ExpDesign()
    surrogate_opts1.ExpDesign.method = 'normal'
    surrogate_opts1.ExpDesign.n_init_samples = 50
    surrogate_opts1.ExpDesign.sampling_method = 'latin-hypercube'
    surrogate1 = surrogate_opts1.create_metamodel()
    print('Surrogate 1 completed')
    print('')

    # Save surrogate
    with open('tsunami0.pk1', 'wb') as output:
        joblib.dump(surrogate0, output, 2)
        
    # Save surrogate
    with open('tsunami1.pk1', 'wb') as output:
        joblib.dump(surrogate1, output, 2)
    
    if 0:
        # Post processing
        L2_PostPCE = bv.PostProcessing(surrogate1)
        L2_PostPCE.plot_moments(plot_type='line')
        # Plot to check validation visually.
        L2_PostPCE.valid_metamodel(n_samples=1)
        # Compute and print RMSE error
        L2_PostPCE.check_accuracy(n_samples=300)
        total_sobol = L2_PostPCE.sobol_indices()
    
    # Test BayesInference
    import copy
    from tsunami_model import tsunami_model
    true_data = tsunami_model(np.array([[90.0,60.0]]))
    model0.observations = true_data
    model1.observations = true_data
    true_data_nox = copy.deepcopy(true_data)
    true_data_nox.pop('x_values')
    
    # Test surrogate output shape
    mean, std = surrogate1.eval_metamodel(np.array([[90.0,60.0]]))
    
    # Bayesian Inference
    if 0:
        # Set uncertainty
        import pandas as pd
        obsData = pd.DataFrame(true_data_nox, columns=model0.Output.names)
        DiscrepancyOpts = bv.Discrepancy('')
        DiscrepancyOpts.type = 'Gaussian'
        DiscrepancyOpts.parameters = (obsData*0.15)**2
        
        # Parameter estimation / single model validation via TOM
        BayesOpts = bv.BayesInference(surrogate1)
        BayesOpts.emulator= True
        BayesOpts.plot_post_pred = True
        #BayesOpts.inference_method = 'rejection'
        BayesOpts.Discrepancy = DiscrepancyOpts
        BayesOpts.bootstrap = True
        BayesOpts.n_bootstrap_itrs = 10
        BayesOpts.bootstrap_noise = 0.05
        BayesOpts.out_dir = ''
        
        # Set the MCMC parameters
        import emcee
        BayesOpts.inference_method = "MCMC"
        BayesOpts.mcmc_params = {
            'n_steps': 1e3,
            'n_walkers': 30,
            'moves': emcee.moves.KDEMove(),
            'multiprocessing': False,
            'verbose': False
            }
         
        Bayes = BayesOpts.create_inference()
        # Pkl the inference results
        with open(f'Bayes_{model0.name}.pkl', 'wb') as output:
            joblib.dump(Bayes, output, 2)
            
    # Model Comparison
    if 1:
        # Set the models
        meta_models = {
            "M0": surrogate0,
            "M1": surrogate1,
            }

        # BME Bootstrap options
        opts_bootstrap = {
            "bootstrap": True,
            "n_samples": 1000,
            "Discrepancy": DiscrepancyOpts,
            "emulator": True,
            "plot_post_pred": False
            }

        # Run model comparison
        BayesOpts = bv.BayesModelComparison(
            justifiability=True,
            n_bootstarp=10,
            just_n_meas=2
            )
        output_dict = BayesOpts.create_model_comparison(
            meta_models,
            opts_bootstrap
            )
                
        import pickle as pkl
        # save model comparison results
        with open('ModelComparison.pkl', 'wb') as f:
            pkl.dump(output_dict, f)
        